-- When timer reaches 0, starts climbing from clinging state
local WALL_THRUST_DELAY = 10
-- For how long player is able to climb without stopping to cling
local WALL_THRUSTING_DURATION = 10

--------------- Hud function implementations ---------------

local function DrawMag(player, x, y, width, height, empty, bac)

    --Slug Repeater 
    --Draw background
    v.drawFill(slug_x, slug_y, slug_width, slug_height, background_mag_color)
    if(stplayer.slug_mag > 0) then
        v.drawFill(slug_x, slug_y, slug_loadedbased_width, slug_height, loaded_mag_color)
    else
        v.drawFill(slug_x, slug_y, slug_emptybased_width, slug_height, empty_mag_color)
    end
end

local function drawweaponhud(v, stplayer, cam)
    
    if(not Valid(stplayer.mo, SKINS.JAZZ) or stplayer.tf_form ~= TF_FORM.BOT_MODE) then
       return false 
    end

    local loaded_mag_color = 135
    local empty_mag_color = 41
    local background_mag_color = 31

    
    -- Adjust to render the bar just below the RINGS HUD
    local slug_x = hudinfo[HUD_RINGS].x
    local slug_y = hudinfo[HUD_RINGS].y + 20
    local slug_height = 3
    local slug_width = 50
    --The ratio that make the bar "move"
    local slug_mag_ratio = FixedDiv(stplayer.slug_mag*FRACUNIT, SLUG_MAG_MAX*FRACUNIT)
    local slug_reload_ratio = FixedDiv(stplayer.slug_reload_tics*FRACUNIT, SLUG_RELOAD_TICS_MAX*FRACUNIT)


    --The adjusted width
    local slug_loadedbased_width = FixedMul(slug_width, slug_mag_ratio)
    local slug_emptybased_width = FixedMul(slug_width, slug_reload_ratio)

    --Slug Repeater
    v.drawFill(slug_x, slug_y, slug_width+1, slug_height+1, background_mag_color)
    if(stplayer.slug_mag > 0) then
        v.drawFill(slug_x, slug_y, slug_loadedbased_width, slug_height, loaded_mag_color)
    else
        v.drawFill(slug_x, slug_y, slug_emptybased_width, slug_height, empty_mag_color)
    end
    
    
    -- Adjust to render the bar just below the RINGS HUD
    local crescent_x = slug_x
    local crescent_y = slug_y + slug_height*2
    local crescent_height = slug_height
    local crescent_width = slug_width
    --The ratio that make the bar "move"
    local crescent_mag_ratio = FixedDiv(stplayer.crescent_mag*FRACUNIT, CRESCENT_MAG_MAX*FRACUNIT)
    local crescent_reload_ratio = FixedDiv(stplayer.crescent_reload_tics*FRACUNIT, CRESCENT_RELOAD_TICS_MAX*FRACUNIT)

    --The adjusted width
    local crescent_loadedbased_width = FixedMul(crescent_width, crescent_mag_ratio)
    local crescent_emptybased_width = FixedMul(crescent_width, crescent_reload_ratio)

    --Crescent Repeater 
    v.drawFill(crescent_x, crescent_y, crescent_width+1, crescent_height+1, background_mag_color)
    if(stplayer.crescent_mag > 0) then
        v.drawFill(crescent_x, crescent_y, crescent_loadedbased_width, crescent_height, loaded_mag_color)
    else
        v.drawFill(crescent_x, crescent_y, crescent_emptybased_width, crescent_height, empty_mag_color)
    end


end


--------------- Hud function definitions ---------------


hud.add(drawweaponhud, "game")


local function SpawnEmptyMag(player)
    
    local icon = P_SpawnMobjFromMobj(player.mo, 0, 0, player.mo.height, MT_EMPT)

    -- print(player.empty_icon)
    --If icon already exists
    if(icon ~= nil and icon.target ~= nil and icon.target.valid == true) then
        P_KillMobj(icon.target)
    end

    icon.target = player.mo
    
end

--Empties the mag
local function EmptyMag(player)
    player.slug_mag = 0 --Empty mag
    player.crescent_mag = 0 --Empty mag
    player.slug_reload_tics = SLUG_RELOAD_TICS_MAX
    player.crescent_reload_tics = CRESCENT_RELOAD_TICS_MAX 
end

---------------------------------------------------------------------
---------- Connecting all of the hooks from this file  --------------
---------------------------------------------------------------------

--Makes the player shoot a single bullet projectile
local function ShootBullet(player)

    player.slug_mag = $-1 --Detect a bullet being shot

    -- Bullet to be shot
    local bullet = P_SpawnMobjFromMobj(player.mo, 0, 0, 30*FRACUNIT, MT_WBUL)
    bullet.scale = player.mo.scale
    bullet.target = player.mo
    bullet.angle = player.mo.angle

    --The total speed that should be distributed to horizontal and vertical speed depending on vertical angle
    local total_speed = mobjinfo[bullet.type].speed
    local side_spread = (player.speed/FRACUNIT)/8 -- Furthest side spread depending on speed
    
    --Forward and vertical speeds
    local bullet_forward_speed  = FixedMul(total_speed, cos(player.aiming))
    local bullet_vertical_speed = FixedMul(total_speed, sin(player.aiming))
    local bullet_side_speed = P_RandomRange(-side_spread, side_spread)*FRACUNIT

    --Shooting the bullet bullet is under water 
    if(bullet.eflags & MFE_UNDERWATER ~= 0) then
        bullet_side_speed = $*2
        bullet_forward_speed = $/2
    end

    -- P_SetObjectMomZ(bullet, bullet_vertical_speed, false)
    P_InstaThrust(bullet, player.mo.angle + ANGLE_90, bullet_side_speed) -- Bullet spread horizontally
    P_SetObjectMomZ(bullet, bullet_vertical_speed, bullet_vertical_speed) -- Side spead
    P_Thrust(bullet, player.mo.angle, bullet_forward_speed) -- Forward movement speed

    --Play souund
    S_StartSound(player.mo, sfx_bullet) -- Bullet sound
end



--Makes the players shoot the heavy blast projectile
local function ShootBlast(player)

    player.crescent_mag = $-1 --Detect a blast being shot
    
    -- blast to be shot
    local blast = P_SpawnMobjFromMobj(player.mo, 0, 0, 30*FRACUNIT, MT_WBLS)
    -- blast.damage = 15
    
    blast.scale = player.mo.scale
    blast.target = player.mo
    blast.angle = player.mo.angle

    --The total speed that should be distributed to horizontal and vertical speed depending on vertical angle
     local total_speed = mobjinfo[blast.type].speed
    local side_spread = (player.speed/FRACUNIT)/8 -- Furthest side spread depending on speed
    
    --Forward and vertical speeds
    local blast_forward_speed  = FixedMul(total_speed, cos(player.aiming))
    local blast_vertical_speed = FixedMul(total_speed, sin(player.aiming))
    local blast_side_speed = P_RandomRange(-side_spread, side_spread)*FRACUNIT

    --Shooting the blast blast is under water 
    if(blast.eflags & MFE_UNDERWATER ~= 0) then
        blast_side_speed = $*2
        blast_forward_speed = $/2
    end

    -- P_SetObjectMomZ(blast, blast_vertical_speed, false)
    P_InstaThrust(blast, player.mo.angle + ANGLE_90, blast_side_speed) -- blast spread horizontally
    P_SetObjectMomZ(blast, blast_vertical_speed, blast_vertical_speed) -- Side spead
    P_Thrust(blast, player.mo.angle, blast_forward_speed) -- Forward movement speed

    --Play souund
    S_StartSound(player.mo, sfx_blast) -- blast sound
end

--Thinker for when the player is in the BulletStance state
local function WeaponStanceThink(player)
    if(player.mo.state == S_WEAPON_STANCE) then
        
        --Makes sure to display correct sprite depending on whether Jazz is airbone or not
        if(not P_IsObjectOnGround(player.mo)) then
            player.mo.sprite2 = SPR2_BUJU
        elseif(P_IsObjectOnGround(player.mo)) then
            player.mo.sprite2 = SPR2_BUSN
            
            --Skid on the ground
            if(player.speed > 0) then
                SpawnOnSide(player, MT_DUST)
            end
        end



        --Hold long is required to hold the shoot button to switch from Slug to Crescent
        local weapon_swap_tic = states[player.mo.state].tics/3

        --Trigger primary and/or secondary weapon
        if(player.spinheld == 0 and player.prevspinheld ~= 0 and player.prevspinheld < weapon_swap_tic) then
            player.burst_shot_interval_tics = BURST_SHOT_INTERVAL_TIC_MAX -- It will trigger shooting bullets since we have several to shoot
        elseif(player.prevspinheld == weapon_swap_tic) then
            --Do we have ammo?
            if(player.crescent_mag > 0 ) then
                ShootBlast(player) -- Single shot
            else
                --The Crescent Mag is EMPTY!!!!!!!
                S_StartSound(player.mo, sfx_empty) 
                SpawnEmptyMag(player)
            end
        end

        --Shoots a single bullet at the time
        if(player.burst_shot_interval_tics ~= 0 and player.burst_shot_interval_tics % 3 == 0) then

            --Check ammo
            if(player.slug_mag > 0) then 
                ShootBullet(player)
            else
                --No ammo, reset
                player.burst_shot_interval_tics = 0
                S_StartSound(player.mo, sfx_empty)
                SpawnEmptyMag(player)
            end
        end
        
        --Counting down the bullet timer 
        if(player.burst_shot_interval_tics > 0) then
            player.burst_shot_interval_tics = $-1
        end
    end    
end


--Thinker that sets the powerup based on the special boost tics
local function PowerUpThink(player)
    --How Jazz is affected in robot mode while under special boost 
    if(player.special_boost_tics > 0) then --Special speed boost is in the affect

        player.normalspeed = skins[player.mo.skin].normalspeed + SPECIAL_BOOST.bot_speed
        player.acceleration = skins[player.mo.skin].acceleration + SPECIAL_BOOST.bot_accel
        local ghost = P_SpawnGhostMobj(player.mo)
        if(ghost ~= nil) then
            ghost.color = SKINCOLOR_SKY -- or CYAN, or any bluish color
            ghost.blendmode = AST_ADD  -- Additive blending gives a glow effect
            ghost.alpha = FRACUNIT/2   -- 50% transparency (adjust as needed)
        end

        -- Set glow effect
        -- player.mo.color = SKINCOLOR_SKY -- or CYAN, or any bluish color
        -- player.mo.renderstyle = RS_TRANS
        -- player.mo.blendmode = AST_ADD  -- Additive blending gives a glow effect
        -- player.mo.alpha = FRACUNIT/2   -- 50% transparency (adjust as needed)

    else --Special speed boost is removed
        player.normalspeed = skins[player.mo.skin].normalspeed
        player.acceleration = skins[player.mo.skin].acceleration

        -- player.mo.renderstyle = RS_NONE
        -- player.mo.blendmode = AST_COPY
        -- player.mo.alpha = FRACUNIT
        player.mo.color = skins[player.mo.skin].prefcolor -- Restore default color
    end

end

local function BotMoveBlocked(pmo, tmo, line)
    
    if line ~= nil and line.valid and pmo.player.jumpheld >= 1 and line.flags&ML_NOCLIMB == 0 then 
        pmo.state = S_PLAY_CLIMB
        pmo.player.jumpspecial = 0
    end
end

--Thinker that sets how the climbing should be handled
local function ClimbingThink(player)

    -- Should we switch to the flipped frame or not?
    player.flip = $ or 0
    
    -- Reduce height of the character while leaping over the wall
    if player.mo.state == S_WALL_LEAP then
        player.mo.height = player.spinheight
    end
    
    -- To make climbing work we need:
    -- 1. trigger to start climbing
    -- 2. logic of climbing
    --
    
    -- Recover from climbing
    local function recover()
        player.wall_bursttimer = 0 
        player.jumpspecial = 0
        AddSpecialBoostTics(player, SPECIAL_BOOST.dmg_tics)
        S_StartSound(player.mo, sfx_jump)
    end

    -- While Climbing
    if player.mo.state == S_PLAY_CLING or player.mo.state == S_PLAY_CLIMB then 
        player.mo.tics = -1
        
         --Information about the wall
        local climb_line = lines[player.lastlinehit]
            
        --Wall x and y points
        local wall_x, wall_y = P_ClosestPointOnLine(player.mo.x, player.mo.y, climb_line)
        --Angle between Jazz position and the closes point on the wall
        local wall_angle = R_PointToAngle2(player.mo.x, player.mo.y, wall_x, wall_y)
        -- --Distance between Jazz and closest point on the walll
        local wall_distance = R_PointToDist2(player.mo.x, player.mo.y, wall_x, wall_y)

        -- Check for jumping over the wall
        local front = climb_line.frontsector
        local back = climb_line.backsector
        local climbing_sector = front
        local other_sector = nil
        local fof_top = nil
        local fof_bottom = nil

        -- -- The actual z coordinates, or platform that we're jumping over
        local effective_jumpover_z = front.ceilingheight 

        local function set_fof(rover)
            if player.mo.z < rover.topheight and player.mo.z > rover.bottomheight then
                effective_jumpover_z = rover.topheight
            end
        end

        -- There's a backsector, player is on opposite side of wall they are climbing
        if(back) then
            other_sector = back
            if P_PointOnLineSide(player.mo.x, player.mo.y, climb_line) == 0 then
                climbing_sector, other_sector = $2, $1
            --     print("player in the front")
            -- else
            --     print("player in the back")
            end

            -- Regular behavior, try to climb up to the floor height
            effective_jumpover_z = climbing_sector.floorheight
            
            -- Special case: Climbing sector appears to be lower than 
            -- the other sector ceiling, so swap and use ceiling
            if climbing_sector.ceilingheight < other_sector.ceilingheight
            and player.mo.z > climbing_sector.ceilingheight then
                climbing_sector, other_sector = $2, $1
                --Irregular behavior, try to climb the ceiling (nothing to jump over)
                effective_jumpover_z = climbing_sector.ceilingheight
                -- print("special case: swap sectors")
            end

            for rover in back.ffloors() do
               set_fof(rover)
            end
        end
        for rover in front.ffloors() do
            set_fof(rover)
        end

        -- Make sure to keep the frames flipped correctly
     


        local function debug_messages()
            print("-----")
            print("-----")
            print("-----")
            print("---- player z: "..(player.mo.z/FRACUNIT))
            print("----- front c: "..(front.ceilingheight/FRACUNIT))
            print("----- front f: "..(front.floorheight/FRACUNIT))
            local floors = 0
            for rover in front.ffloors() do
                print("---front fof t: "..rover.topheight/FRACUNIT)
                print("---front fof b: "..rover.bottomheight/FRACUNIT)
                floors = $+1
            end
            
            if(back) then
                print("---------- player z: "..(player.mo.z/FRACUNIT))
                print("------------ back c: "..back.ceilingheight/FRACUNIT)
                print("------------ back f: "..back.floorheight/FRACUNIT)
                for rover in back.ffloors() do
                    print("----back fof t: "..rover.topheight/FRACUNIT)
                    print("----back fof b: "..rover.bottomheight/FRACUNIT)
                    floors = $+1
                end
                
            end
            if(climbing_sector == front) then
                print("> climbing front")
            elseif climbing_sector == back then
                print("> climbing back")
            end
            print("------------- fof's  "..floors)
            print("== effective_jumpover_z: "..effective_jumpover_z/FRACUNIT)
        end
        -- debug_messages()

        -- print(tostring(P_CheckPosition(player.mo, 
        -- FixedMul(cos(wall_angle), 20*FRACUNIT),
        -- FixedMul(sin(wall_angle), 20*FRACUNIT))))
        -- print(tostring(P_LineIsBlocking(player.mo, climb_line)))

        local function wall_thrust_thinker() 
            local pmo = player.mo
            
            if(player.wall_bursttimer == nil) then
                player.wall_bursttimer = 0
            end
            if(player.wall_durationtimer == nil) then
                player.wall_durationtimer = 0
            end
            player.wall_durationtimer = $ or 0

            -- Countdown burst timer
            if player.mo.state == S_PLAY_CLING and player.wall_bursttimer > 0 then
                player.wall_bursttimer = $ - 1
            end
            -- Countdown duration timer
            if player.mo.state == S_PLAY_CLIMB and player.wall_durationtimer > 0 then
                player.wall_durationtimer = $ - 1
            end

            -- Cling if the duration of the climbing has exceeded
            if player.mo.state == S_PLAY_CLIMB and player.wall_durationtimer <= 0 then
                player.mo.state = S_PLAY_CLING
                if player.flip == 1 then
                    player.flip = 0
                else
                    player.flip = 1
                end
                player.wall_durationtimer = WALL_THRUSTING_DURATION
                player.mo.momz = $/2 -- Slow down when clinging
                
                local crack_z = player.mo.height
                local crack_horiz = ANGLE_90
                if player.flip == 1 then
                    crack_horiz = InvAngle(crack_horiz)
                end
                local side_angle = wall_angle+crack_horiz
                local crack = P_SpawnMobjFromMobj(player.mo, 
                                                FixedMul(cos(side_angle), player.mo.radius) + FixedMul(cos(wall_angle), wall_distance/2),
                                                FixedMul(sin(side_angle), player.mo.radius) + FixedMul(sin(wall_angle), wall_distance/2),
                                                 crack_z, MT_CRCK)

                crack.angle = wall_angle
                -- P_InstaThrust(crack, (-player.flip) * wall_angle - ANGLE_90, crack_horiz)
            end

            -- Perform a thrust upward
            if player.cmd.forwardmove and player.wall_bursttimer <= 0 then
                -- Burst upward
                -- pmo.momz = 5 * FRACUNIT
                P_SetObjectMomZ(pmo, 15*FRACUNIT, false)
                S_StartSound(player.mo, sfx_clmb_a)
                player.wall_bursttimer = WALL_THRUST_DELAY
                player.mo.state = S_PLAY_CLIMB

                return true
            end

            return false
        end
        
        -- If player doesn't jump out of the climbing state, continue climbing
        wall_thrust_thinker()
        if(player.mo.momz <= 0) then -- Detect continues clinging
            P_SetObjectMomZ(player.mo, 0, false)
            player.mo.state = S_PLAY_CLING
        end
        P_InstaThrust(player.mo, player.mo.angle, 0) -- Keep same horizontal position
        
        if player.flip == 1 then
            player.mo.frame = 0
        else
            player.mo.frame = 1
        end

        -- Jazz has climbed above the wall
        if player.mo.z >= effective_jumpover_z then
            -- You can then push them up and onto the ledge
            player.mo.state = S_WALL_LEAP
            P_SetObjectMomZ(player.mo, 5*FRACUNIT, false)
            P_InstaThrust(player.mo, wall_angle, 5*FRACUNIT)
            recover()
        elseif(player.jumpheld == 1) then -- If player jumps out of the climbing state
            recover()
            player.mo.state = S_PLAY_JUMP
            P_InstaThrust(player.mo, wall_angle + ANGLE_180, 10*FRACUNIT) 
       end

    end
end


local function BotPlayerThink(player)

    
    --Landing effect
    if(player.mo.eflags & MFE_JUSTHITFLOOR and not S_SoundPlaying(player.mo, sfx_land_a)) then
        S_StartSound(player.mo, sfx_land_a)
        SpawnOnSide(player, MT_SPINDUST)
    end
    
    --If taking damage, then emptying the mag
    if(player.mo.state == mobjinfo[player.mo.type].painstate) then
       EmptyMag(player)
    end

    -- If Jazz is active
    if(Active(player)) then

        --Start shooting the Bullet projectile
        if(player.spinheld == 1) then
            player.mo.state = S_WEAPON_STANCE
        end

        --Execute the weapon stance thinker
        WeaponStanceThink(player)

        --Execute the power up thinker
        PowerUpThink(player)
        
        --Execute the climbing thinker
        ClimbingThink(player)
        
        if(player.mo.state ~= S_POUND_FALL and player.jumpheld >= 5 and player.jumpspecial == 1) then
            player.mo.state = S_POUND_FALL
            -- P_SetObjectMomZ(player.mo, -5*FRACUNIT, true)
        end

        if(player.mo.state == S_POUND_GROUND) then
            player.pflags = $|PF_JUMPSTASIS
        end

        
    end
end

local function BlastMobjThink(blast)
    
    local ghost = P_SpawnGhostMobj(blast)
    
    -- local scaling = FRACUNIT
    if(blast.state == mobjinfo[blast.type].deathstate) then
        
        ghost.scale = blast.scale-2*FRACUNIT
        
    end
    
end

local function BlastMoveCollide(blast, collide_mo) 
end

local function BulletMobjThinker(bullet)
    --Underwater spread of the bullet
    if(bullet.eflags & MFE_UNDERWATER) then
        bullet.momx = 9*$/10
        bullet.momy = 9*$/10
    end
    P_SpawnGhostMobj(bullet)
end

local function BulletMoveBlocked(bullet, enemy, wall)
    -- Bullet flies into the wall
    if(wall ~= nil) then 
        S_StartSound(wall.frontsector, sfx_miss)    
    -- Bullet hits the entity
    else
        S_StartSound(enemy, sfx_hit)
    end
end

addHook("MobjThinker", function (iconmo)
    -- if(iconmo.valid == true and iconmo.player ~= nil and iconmo.player.valid == true) then
    if(iconmo.target ~= nil and iconmo.target.valid == true) then
        P_MoveOrigin(iconmo, iconmo.target.x, iconmo.target.y, iconmo.target.z + iconmo.target.height)
    else
        P_KillMobj(iconmo)
    end
end, MT_EMPT)

addHook("PlayerThink", function(player) 
    if(Valid(player.mo, SKINS.JAZZ) and player.tf_form == TF_FORM.BOT_MODE) then
        BotPlayerThink(player)
    end
end)

addHook("MobjThinker", function(mo)
    BulletMobjThinker(mo)
end, MT_WBUL)


addHook("MobjMoveBlocked", function(mo, tmmo, line)
    BotMoveBlocked(mo, tmmo, line)
end, MT_PLAYER)


addHook("MobjMoveBlocked", function(mo, tmmo, line)
    BulletMoveBlocked(mo, tmmo, line)
end, MT_WBUL)

addHook("MobjThinker", function(mo)
    BlastMobjThink(mo)
end, MT_WBLS)

addHook("MobjMoveCollide", function(mo, collide_mo)
    BlastMoveCollide(mo, collide_mo)
end, MT_WBLS)

addHook("ShouldDamage", function(target, inflictor, source, damage, damagetype) 
    -- if(inflictor.type == MT_WBLS and source.state == S_BLAST_SHOT) then
    if(inflictor ~= nil and inflictor.valid and inflictor.type == MT_WBLS and damage == 1) then
        P_DamageMobj(target, inflictor, source, inflictor.info.damage)
        return false
    end
end)

