
---Force applied every tic
local CAR_ACCELERATION = 5*FRACUNIT/2 
---Force applied every tic while using speed sneakers
local CAR_ACCELERATION_BOOSTED = CAR_ACCELERATION*2 
---Fastest speed while driving normally
local SPEED_LIMIT = 100*FRACUNIT 
---Fastest speed while driving while using speed sneakers
local SPEED_LIMIT_BOOSTED = 3*SPEED_LIMIT/2 
---Up to what speed the engine is in the idle state
local IDLE_SPEED = 2*FRACUNIT 
---Up to what speed the engine is in the middle speed state
local MID_SPEED = 2*SPEED_LIMIT/3 
---The turning angle of steering at minimum speed and maximum speeds
local STEERING_LIMIT_ANGLE = {MAX = ANG1*7, MIN = ANG1*2} 
---Fastest speed while driving normally
local REVERSE_SPEED_LIMIT = 10*FRACUNIT 


--------------- Hud function implementations ---------------

local function drawgears(v, stplayer, cam)
    
    if(not Valid(stplayer.mo, SKINS.JAZZ) or stplayer.tf_form ~= TF_FORM.ALT_MODE) then
       return false 
    end

    local gear_drive = v.cachePatch("GEAR_DRIVE")
    local gear_park = v.cachePatch("GEAR_PARK")
    local gear_reverse = v.cachePatch("GEAR_REVERSE")
    local gear_none = v.cachePatch("GEAR_NONE")

    local gear_to_draw = nil


    if(stplayer.mo.car_gear == GEAR_MODE.DRIVE) then
        gear_to_draw = gear_drive
    elseif(stplayer.mo.car_gear == GEAR_MODE.PARK) then
        gear_to_draw = gear_park
    elseif(stplayer.mo.car_gear == GEAR_MODE.REVERSE) then
        gear_to_draw = gear_reverse
    else
        gear_to_draw = gear_none
    end

    local xbarsize, ybarsize = 32*FRACUNIT, 32*FRACUNIT

    local x, y = 17*FRACUNIT, 55*FRACUNIT

    local xscale, yscale = FRACUNIT, FRACUNIT

    local background_gear_color = 31

    v.drawFill(x/FRACUNIT, y/FRACUNIT, gear_to_draw.width+3, gear_to_draw.height+2, background_gear_color)
    v.drawStretched(x, y, xscale, yscale, gear_to_draw)
    -- v.drawCropped(x, y, xscale, yscale, fullbar, 0, nil, 0, 0, FixedMul(FixedDiv(stplayer.mo.phasestamina, stplayer.mo.max_phasestamina), xbarsize), ybarsize)
end


--------------- Hud function definitions ---------------


hud.add(drawgears, "game")


---------------------------------------------------------------------
---------------- Functions and hook definitions  --------------------
---------------------------------------------------------------------

-- local function rotateTheta(theta, delta, theta_prime) 
--     -- local increment = FixedDiv(AngleFixed(theta_prime), AngleFixed(theta))
    
--     -- local new_theta_value = FixedMul(increment, theta) / delta

--     -- return new_theta_value



-- end






--[[
    1. There's a 'moving angle' in which direction gas is applied, sprite always faces that direction
    2. Player's 'actual angle' will always be set to some value to prevent look/turn inputs mouse/joystick, 
it would be set to gradually move towards the moveing angle
    3. Steering would be be changing 'moving angle'
]]--

local function PontiacFlameThink(player)    
    if(player.speed > BREAK_SPEED or player.special_boost_tics > 0) then
        
        local mo = player.mo
        -- Create the follower object once
        if not player.follower or not player.follower.valid then
            player.follower = P_SpawnMobj(mo.x, mo.y, mo.z, MT_POFL) -- Replace with your custom mobj type
            -- player.follower.flags2 = $|MF2_DONTDRAW -- Optional: remove this if you want it visible
        end

        -- How far in front of the player (fixed-point units)
        local distance = player.mo.radius

        -- Angle math to get in-front position
        local angle = mo.car_forward_angle
        local x = mo.x + FixedMul(cos(angle), distance)
        local y = mo.y + FixedMul(sin(angle), distance)
        local z = mo.z + mo.height/2 -- Keep it around chest height

        -- Update the follower's position
        local follower = player.follower
        P_MoveOrigin(follower, x, y, z)
        follower.angle = angle -- Match player's direction if needed
    elseif(player.follower ~= nil and player.follower.valid) then
        P_KillMobj(player.follower)
    end
end


--Returns whether the car can bump the collided into object
local function CheckBump(car_mo, collide_mo)
    return (car_mo.player.speed > BUMP_SPEED or car_mo.car_traction == CAR_TRACTION.OVERSTEER_CONTROL) 
    and collide_mo.flags & BUMP_MF and L_ZCollide(car_mo, collide_mo)
end


--Returns the FRACUNIT difference with signed direction of two angles given as FRACUNITS
--Direction is -1 for anticlockwise rotation from start angle to finish angle, and 1 for clockwise
local function AngleDiff(fixed_start, fixed_finish)

    -- local fixed_start = AngleFixed(player.mo.angle)
    -- local fixed_finish = AngleFixed(player.mo.car_forward_angle)

    local raw_diff = fixed_start - fixed_finish --Moving angle - player angle
    local final_diff = abs(raw_diff)
    local direction = -1
    if(raw_diff < 0) then
        direction = $*-1
    end

    if(final_diff > 180*FRACUNIT) then
        final_diff = abs($-360*FRACUNIT) --Making sure we get the shortest path between started and finished angle
        direction = $*-1
    end

    return final_diff * direction
end


--Tries to apply as many tire effects as it can
local function TireEffects(player)
  
    local car_accel  = FixedMul(CAR_ACCELERATION, player.mo.scale)
    local idle_speed = FixedMul(IDLE_SPEED, player.mo.scale)

    if(P_IsObjectOnGround(player.mo)) then
        local side_offset = player.mo.radius*FRACUNIT
        if(player.mo.car_accel > idle_speed) then
            SpawnOnSide(player, MT_TRCK, side_offset)    
        end


        if(player.mo.car_accel > car_accel) then 
            SpawnOnSide(player, MT_FIRE, side_offset)
        end

        --The car is oversteering
        if(player.mo.car_traction == CAR_TRACTION.OVERSTEER_CONTROL) then

            --Drifting effects
            -- if(player.car_brake_input >= 1 and player.car_gas_input >= 1) then
            if(player.speed > idle_speed) then
                SpawnOnSide(player, MT_SPINDUST, side_offset)
            end

        end 
    end
end

--Tries to apply as many car sounds as it can
local function CarSounds(player)

    --Using the horn
    if(input.gameControlDown(GC_CUSTOM3) and not S_SoundPlaying(player.mo, sfx_horn)) then
        S_StartSound(player.mo, sfx_horn)
    end


    --Starting the engine
    if(Active(player) and not S_SoundPlaying(player.mo, sfx_rev) and player.speed < IDLE_SPEED and player.mo.car_accel >= 1) then
        S_StartSound(player.mo, sfx_rev)
    end

    
    --Idle working engine sound
    if(player.sound_tics_idle <= 0 and player.mo.car_gear == GEAR_MODE.PARK) then
        S_StartSound(player.mo, sfx_idle)
        player.sound_tics_idle = MAX_IDLE_SOUND_TICS
    else
        
    end
    

    --Mid working engine sound
    if(player.mo.car_gear ~= GEAR_MODE.PARK and player.sound_tics_mid <= 0 and player.speed <= MID_SPEED and player.speed > IDLE_SPEED) then
        S_StartSound(player.mo, sfx_mid)
        player.sound_tics_mid = MAX_MID_SOUND_TICS
        S_StopSoundByID(player.mo, sfx_high)
    else
    end

    --High working engine sound
    if(player.mo.car_gear ~= GEAR_MODE.PARK and player.sound_tics_high <= 0 and player.speed > MID_SPEED) then
        S_StartSound(player.mo, sfx_high)
        player.sound_tics_high = MAX_HIGH_SOUND_TICS
        S_StopSoundByID(player.mo, sfx_mid)
    else
        
    end

    --Oversteering
    if(player.mo.car_traction == CAR_TRACTION.OVERSTEER_CONTROL) then
        -- Braking sound
        if(not S_SoundPlaying(player.mo, sfx_brake)) then
            S_StartSound(player.mo, sfx_brake) 
        end
    else
        --Stop braking sounds
        if(S_SoundPlaying(player.mo, sfx_brake)) then
            S_StopSoundByID(player.mo, sfx_brake)
        end
    end

    player.sound_tics_idle = $-1
    player.sound_tics_mid = $-1
    player.sound_tics_high = $-1

end

--Restricts undesired input, as well as gathering needed input for turning and moving the car
local function AffectInput(player)

    ------------------------------------------------
    ------------------------------------------------  RESTRICT REGULAR INPUT
    ------------------------------------------------

    --Cancel jumping and regular movement
    player.pflags = $|PF_FULLSTASIS
    
    ------------------------------------------------
    ------------------------------------------------  GATHER STEERING INPUT
    ------------------------------------------------
    
    --Side input between -1 and 1
    player.car_steer_input = nil

    local side_input_keyboard = nil
    local side_input_joystick = nil
    if(player == secondarydisplayplayer) then
        side_input_keyboard = (input.gameControl2Down(GC_STRAFELEFT) - input.gameControl2Down(GC_STRAFERIGHT)) * 1000
        side_input_joystick = -clamp(input.joy2Axis(JA_STRAFE), -1000, 1000)
    else    
        --Retrieves the left and right steering 
        side_input_keyboard = (input.gameControlDown(GC_STRAFELEFT) - input.gameControlDown(GC_STRAFERIGHT)) * 1000
        side_input_joystick = -clamp(input.joyAxis(JA_STRAFE), -1000, 1000)
    end

    local side_input = 0
    
    -- Choosing either the keyboard or joystick input for pitch
    if(abs(side_input_joystick) >= abs(side_input_keyboard)) then
        side_input = side_input_joystick 
    else 
        side_input = side_input_keyboard             
    end

    player.car_steer_input = FRACUNIT/1000 * side_input

    ------------------------------------------------
    ------------------------------------------------  GATHER MOVEMENT INPUT
    ------------------------------------------------

    if(player == secondarydisplayplayer) then
        -- Car movement inputs (It should be part of player to always know gas and brake input)
        player.car_gas_input =   max(input.gameControl2Down(GC_FORWARD) and 1 or 0, input.gameControl2Down(GC_FIRE) and 1 or 0)
        player.car_brake_input = max(player.spinheld > 0 and 1 or 0, max(input.gameControl2Down(GC_BACKWARD) and 1 or 0,  input.gameControl2Down(GC_FIRENORMAL) and 1 or 0))
    else
        -- Car movement inputs (It should be part of player to always know gas and brake input)
        player.car_gas_input =   max(input.gameControlDown(GC_FORWARD) and 1 or 0, input.gameControlDown(GC_FIRE) and 1 or 0)
        player.car_brake_input = max(player.spinheld > 0 and 1 or 0, max(input.gameControlDown(GC_BACKWARD) and 1 or 0,  input.gameControlDown(GC_FIRENORMAL) and 1 or 0))
    end

    --Make sure the car faces the spring direction
    if(player.powers[pw_justsprung] ~= 0) then
        player.mo.car_forward_angle = player.mo.angle
    end
        
    
end



--Uses the gathered input to turn left and right
local function ApplySteering(player)

    --[[
    --Steering boundries 
    local max_steering_speed = nil 
    -- Change some behaviors while braking
    if(player.car_brake_input >= 1) then
        -- Increase the steering angle to simulate "drifting"
        max_steering_speed = DRIFTING_STEERING_ANGLE
    else
        max_steering_speed = REGULAR_STEERING_ANGLE
    end
    ]]--

       
    ------------------------------------------------
    ------------------------------------------------  SETS THE ACTUAL TURNING AND SPRITE ROTATION
    ------------------------------------------------

    --The angle of the turn due to steering
    local turn_angle = 0

    --Scaled limits
    local speed_limit = FixedMul(SPEED_LIMIT, player.mo.scale)
    local idle_speed = FixedMul(IDLE_SPEED, player.mo.scale)


    --Only turn when moving
    if(player.mo.car_gear ~= GEAR_MODE.PARK) then
        
        if(player.mo.car_traction == CAR_TRACTION.NO_CONTROL) then -- No control, allow for minimal turn
            turn_angle = STEERING_LIMIT_ANGLE.MIN
            -- print("No control: "..turn_angle/ANG1)
        elseif(player.mo.car_traction == CAR_TRACTION.OVERSTEER_CONTROL) then -- Halted control, thus too much turn
            turn_angle = STEERING_LIMIT_ANGLE.MAX
            -- print("Oversteer: "..turn_angle/ANG1)
        -- elseif(player.mo.car_traction == CAR_TRACTION.IN_CONTROL) then -- Full control, thus normal turn
        else
            --Ratio of current speed and maximum speed to affect turning 
            local turn_speed = clamp(FRACUNIT - FixedDiv(player.speed, speed_limit), 0, FRACUNIT)
            --Calculate the actual angle of the turn based on steering boundries and how fast we're going
            turn_angle = ease.linear(turn_speed, STEERING_LIMIT_ANGLE.MIN, STEERING_LIMIT_ANGLE.MAX)
            -- print("Normal: "..turn_angle/ANG1)
        end
        
        local steering_strength = ease.linear(player.car_steer_input, 0, turn_angle)

        --Changse the angle of the car
        player.mo.car_forward_angle = $ + clamp(steering_strength, -turn_angle, turn_angle)
        player.drawangle = player.mo.car_forward_angle -- Make sprites face the movement angle
    end
    
   
    ------------------------------------------------
    ------------------------------------------------  AFFECTS PLAYER'S ANGLE PURELY FOR CAMERA CHANGING PURPOSES
    ------------------------------------------------

    --Playstyle flag
    local automatic_flags = PF_ANALOGMODE|PF_DIRECTIONCHAR
    local desired_angle = player.mo.car_forward_angle
    -- if(player.speed < 5*FRACUNIT) then -- Don't rotate!
    --         print(player.mo.car_forward_angle)    
    if(player.mo.car_gear == GEAR_MODE.REVERSE) then
        desired_angle = $ + ANGLE_180
    end

    

    --Turn the actual character only if not in automatic playstyle mode
    if(player.pflags & automatic_flags ~= automatic_flags) then
        local final_diff = AngleDiff(AngleFixed(player.mo.angle), AngleFixed(desired_angle))
        -- print(final_diff/FRACUNIT)
        
        
        if(player.car_gear == GEAR_MODE.DRIVE and abs(final_diff) > 90*FRACUNIT) then -- Detect forced big change of the angle by exterior sources such as horizontal springs
            player.mo.car_forward_angle = player.mo.angle
        else -- Make player's angle follow car's actual angle
            player.mo.angle = $ + ease.inquad(FRACUNIT/2, 0, FixedAngle(final_diff))
            -- if(player.mo.car_gear == GEAR_MODE.REVERSE) then
            --     player.mo.angle = $ - ANGLE_180
            -- end
        end

    end


end

--Uses the gathered input to apply the gas and brake 
local function ApplyMove(player)


    ------------------------------------------------
    ------------------------------------------------  LIST OF ALL VARIABLES DEFINED IN THE FUNCTION 
    ------------------------------------------------
    

    --The maximum applied FRACUNIT acceleration
    local max_acceleration = nil
    --The maximum FRACUNIT speed 
    local max_speed = nil
    --FRACUNIT Momentum direction of the car
    local movement_angle = nil
    --FRACUNIT force or speed brakes if applied
    local brake_power = nil
    --Player's desired FRACUNIT force to move the car forward
    local forward_force = 0
    --Player's desired FRACUNIT force that stops the car's movement
    local braking_force = 0
    --Total FRACUNIT acceleration to be applied every tic
    local desired_accel = nil
    --Forward and backward FRACUNIT speed (positive and negative respectively)
    local signed_speed = nil
    --Actual FRACUNIT friction fo the surface 
    local sector_friction = nil
    --Actual FRACUNIT friction applied on the car (since tires reduce friction DUH)
    local applied_friction = nil
    

    --Don't move while parked
    if(player.mo.car_gear == GEAR_MODE.PARK) then --When parked
        player.pflags = $|PF_STARTDASH --Charge spin dash so that we can unlock spindash related areas
        player.mo.car_accel = 0 -- Reset the speed coming from the car
    else
        ------------------------------------------------
        ------------------------------------------------  DEFINE REGULAR BOUNDRIES AND SPECIAL VALUES
        ------------------------------------------------

        player.pflags = $&~PF_STARTDASH

        --Applying regular movement
        max_acceleration = FixedMul(CAR_ACCELERATION, player.mo.scale)
        max_speed = FixedMul(SPEED_LIMIT, player.mo.scale)


        --Scaled limit
        local idle_speed = FixedMul(IDLE_SPEED, player.mo.scale)
        local reverse_speed = FixedMul(REVERSE_SPEED_LIMIT, player.mo.scale)

        --Define how strong the brakes are if used
        brake_power = FixedMul(3*max_acceleration/5, player.mo.scale)

        if(player.mo.car_gear == GEAR_MODE.DRIVE) then
            --Adding speed sneakers
            if(player.powers[pw_sneakers] ~= 0) then
                max_acceleration = FixedMul(CAR_ACCELERATION_BOOSTED, player.mo.scale)
                max_speed = FixedMul(SPEED_LIMIT_BOOSTED, player.mo.scale)
            end
            --Adding special boost to maximum acceleration and speeds
            if(player.special_boost_tics > 0) then --Special speed boost is in the affect
                max_speed = $ + FixedMul(SPECIAL_BOOST.car_speed, player.mo.scale)
                max_acceleration = $ + FixedMul(SPECIAL_BOOST.car_accel, player.mo.scale)
            end
        elseif(player.mo.car_gear == GEAR_MODE.REVERSE) then
            max_acceleration = 9*$/10
            max_speed = reverse_speed
        end
            

        if(player.mo.car_traction == CAR_TRACTION.NO_CONTROL) then -- No control, allow for bare minimum movement
            max_acceleration = $/5
            max_speed = $/5
        -- elseif(player.mo.car_traction == CAR_TRACTION.NO_CONTROL) then -- Oversteering, allow for medium control
        --     max_acceleration = $/2
        --     max_speed = $/2 
        end


        --Change friction of the car
        sector_friction = player.mo.subsector.sector.friction
        if(player.mo.car_gear == GEAR_MODE.PARK or (player.car_gas_input == 0)) then
            applied_friction = sector_friction+FRACUNIT/30
        else

            applied_friction = sector_friction+FRACUNIT/20
        end


        --Friction for the car
        player.mo.friction = clamp(applied_friction, sector_friction, FRACUNIT-1)

        
        --Forward and backward speed (positive and negative respectively)
        signed_speed = FixedMul(player.mo.momx, cos(player.mo.car_forward_angle)) + FixedMul(player.mo.momy, sin(player.mo.car_forward_angle))


        ------------------------------------------------
        ------------------------------------------------  ANGLE MATH
        ------------------------------------------------
        
        --Momentum direction of the car
        -- movement_angle = R_PointToAngle2(0, 0, player.mo.momx, player.mo.momy)

        -- angle_dif = AngleDiff(AngleFixed(movement_angle), AngleFixed(player.mo.car_forward_angle))


        ------------------------------------------------
        ------------------------------------------------  DEFINE DEFAULT APPLIED MOMENTUM VARIABLES UNDER "PERFECT CONDITIONS"
        ------------------------------------------------
        
        if(player.mo.car_gear == GEAR_MODE.DRIVE) then
            forward_force = player.car_gas_input*max_acceleration --Your gas petal
            braking_force = player.car_brake_input*brake_power    --You brake petal
        elseif(player.mo.car_gear == GEAR_MODE.REVERSE) then
            forward_force = -player.car_gas_input*max_acceleration --Your gas petal
            -- braking_force = player.car_brake_input*brake_power    --You brake petal
        end
        
        

        --Make sure brakes won't make you drive backwards
        if(signed_speed <= 0) then
            braking_force = 0
        end

        desired_accel = forward_force - braking_force   --Result

        --Limiting the maximum speed 
        if(player.speed > max_speed) then
            desired_accel = 0
        end


        --Making sure that autobrake feature woudn't slow us down while driving
        if(abs(desired_accel) > idle_speed or not P_IsObjectOnGround(player.mo)) then
            player.cmd.forwardmove = 1
        end

        ------------------------------------------------
        ------------------------------------------------  INTRODUCE "INPERFECT CONDITIONS" FOR MOMENTUM VALUES MENTIONED PREVIOUSLY
        ------------------------------------------------

        
        ------------------------------------------------
        ------------------------------------------------  APPLY THE RESULT
        ------------------------------------------------
        
        
        --Final acceeleration is set
        player.mo.car_accel = desired_accel

        --Apply the final movement
        P_Thrust(player.mo, player.mo.car_forward_angle, player.mo.car_accel)
    end
end


--Decides the gear type of the car automatically
local function DecideGear(player)
    local idle_speed = FixedMul(IDLE_SPEED, player.mo.scale)

    --While holding brake: reverse
    --While standing and not holding forward: park
    
    if(player.mo.car_gear == GEAR_MODE.PARK and player.car_brake_input > 0) then
        player.mo.car_gear = GEAR_MODE.REVERSE 
    elseif(player.mo.car_gear == GEAR_MODE.PARK and player.car_gas_input > 0) then
        player.mo.car_gear = GEAR_MODE.DRIVE 
    elseif((player.mo.car_gear == GEAR_MODE.DRIVE and player.speed <= idle_speed and player.car_gas_input == 0)
    or (player.car_brake_input == 0 and player.mo.car_gear == GEAR_MODE.REVERSE)) then
        player.mo.car_gear = GEAR_MODE.PARK
    end
            
end

--Decides the grip type of the car tires on the ground 
local function DecideGrip(player)
    --Momentum direction of the car
    local movement_angle = R_PointToAngle2(0, 0, player.mo.momx, player.mo.momy)
    --Whereis the car wants to move?
    local forward_angle = player.mo.car_forward_angle 
    --Signed FRACUNIT difference between desired movement angle and actual movement of the car
    local move_diff = 0
    
    --If not moving, make the movement default to the direction the car is facing
    if(player.speed == 0) then
        movement_angle = player.mo.car_forward_angle
    end


    --Moving backward
    if(player.speed ~= 0 and player.mo.car_gear == GEAR_MODE.REVERSE) then
        forward_angle = $ + ANGLE_180
    end

    --Difference between car's desired movement and its actual movement in 3D space
    move_diff = AngleDiff(AngleFixed(movement_angle), AngleFixed(forward_angle))

    --Decide traction/grip type
    if(not P_IsObjectOnGround(player.mo) --In the air
    or player.mo.eflags & MFE_UNDERWATER
    or player.mo.eflags & MFE_GOOWATER
    or player.mo.eflags & MFE_TOUCHLAVA) then --Underwater
        player.mo.car_traction = CAR_TRACTION.NO_CONTROL
    elseif((abs(move_diff) > GRIP_LIMIT --You've actually oversteering
    or (player.mo.car_gear ~= GEAR_MODE.REVERSE and player.car_gas_input > 0 and player.car_brake_input > 0)) --or while braking and pressing gas
    and player.mo.car_gear ~= GEAR_MODE.PARK) then --But never oversteer while parking
        player.mo.car_traction = CAR_TRACTION.OVERSTEER_CONTROL
    else 
        player.mo.car_traction = CAR_TRACTION.IN_CONTROL
    end
end


--Be able to break obstacles and power up springs by reaching a certain speed limit
local function PowerUp(player, limit)
    --Break speed
    local speed = GetRealSpeed(player)
    
    --Bust objects in car mode
    if(speed > limit or player.special_boost_tics > 0) then
        player.powers[pw_strong] = $|PONTIAC_POWERUP_FLAGS
    else
        player.powers[pw_strong] = $&~PONTIAC_POWERUP_FLAGS
    end

end


local function CarPreThink(player)

    --Retrieves and limits players input
    AffectInput(player)

    --Displays tire effects
    if(player.mo.state == S_PONTIAC_MODE) then
        TireEffects(player)
    end

    --Executes the car sounds
    CarSounds(player)
end


local function CarPlayerThink(player)

    --Supplies the current grip
    DecideGrip(player)

    DecideGear(player)

    

    if(Active(player)) then
        --Turns the car
        ApplySteering(player)

        -- Car's actual movement
        ApplyMove(player)

        PowerUp(player, BREAK_SPEED)
    end

end


local function CarPostThink(player)
    -- if(player.mo.state ~= mobjinfo[player.mo.type].painstate) then
    
    
    PontiacFlameThink(player)

end


local function CarMobjMoveCollide(car_mo, collide_mo)

    -- Collide with the object only if it meets criteria
    if(CheckBump(car_mo, collide_mo)) then
        S_StartSound(car_mo, sfx_bump)
        
        --Destroying the monitors
        if(collide_mo.flags & MF_MONITOR) then
            P_KillMobj(collide_mo, car_mo, car_mo)
         
        --Getting enemies out of the way
        elseif(collide_mo.flags & MF_ENEMY) then    
            -- Bumping objects out of the way
            P_SetObjectMomZ(collide_mo, FRACUNIT*10, false)
            P_InstaThrust(collide_mo, R_PointToAngle2(0, 0, car_mo.momx, car_mo.momy), -FRACUNIT*5)
            if(car_mo.player.speed > BREAK_SPEED or car_mo.player.special_boost_tics > 0) then
                P_DamageMobj(collide_mo, car_mo, car_mo, 1)
            end
        end
       
    end
end


local function CarMobjDamage(car_mo, inflictor, source, damage, damagetype)
    --Evading damage while bumping the enemy
    -- if (inflictor == source and damage == 1 -- If light damage if source damaged us
    -- or (source ~= nil and source & BUMP_MF)) then  -- If source is bumpable
    if(inflictor ~= nil and inflictor == source and CheckBump(car_mo, inflictor)) then
        return true --Evade damage
    else -- If not evading damage therwise
        if(not S_SoundPlaying(car_mo, sfx_horn)) then
            S_StartSound(car_mo, sfx_horn)
        end
    end
end


---------------------------------------------------------------------
---------- Connecting all of the hooks from this file  --------------
---------------------------------------------------------------------


addHook("PlayerThink", function(player) 
    if(Valid(player.mo, SKINS.JAZZ) and (player.tf_form == TF_FORM.ALT_MODE or player.tf_form == TF_FORM.TF_TO_ALT_MODE)) then
        CarPlayerThink(player) 
    end
end)



--The Base Thinker that plays before others,
--Mostly used to record players input
addHook("PreThinkFrame", function()
	for player in players.iterate() do
        if(Valid(player.mo, SKINS.JAZZ) and (player.tf_form == TF_FORM.ALT_MODE or player.tf_form == TF_FORM.TF_TO_ALT_MODE)) then
            CarPreThink(player)
        end
    end
end)


--The Base Thinker that plays before others,
--Mostly used to record players input
addHook("PostThinkFrame", function()
	for player in players.iterate() do
        if(Valid(player.mo, SKINS.JAZZ) and player.tf_form == TF_FORM.ALT_MODE) then
            CarPostThink(player)
        end
    end
end)

--Handles collision of the car mode with the objects
addHook("MobjMoveCollide", function(mo, collide_mo)
    if(Valid(mo, SKINS.JAZZ) and collide_mo ~= nil and mo.player.tf_form == TF_FORM.ALT_MODE) then
        CarMobjMoveCollide(mo, collide_mo)
    end
end, MT_PLAYER)


addHook("MobjDamage", function(target, inflictor, source, damage, damagetype)
    
    if(Valid(target, SKINS.JAZZ) and target.player.tf_form == TF_FORM.ALT_MODE) then
        return CarMobjDamage(target, inflictor, source, damage, damagetype)
    end
end, MT_PLAYER)


